# The author program imports a list of names (author_config.txt)
# These participants can vote on who will be the 1th, 2th, etc.
# author one some paper. This is done via a graphical interface
# and it is not possible to vote for yourself.
#
# Note that if two (or more) players receive the same amount of
# points, the bitonic sort algorithm will will provide a default
# solution. The person nearest the bottom of the
# author_config.txt file will be ranked first.
#
# The implementation is based on VIFF, the Virtual Ideal
# Functionality Framework, while the GUI toolkit used is PyGTK.

from twisted.internet import gtk2reactor # For gtk-2.0
gtk2reactor.install()
from twisted.internet import reactor

from optparse import OptionParser
from viff.field import GF
from viff.runtime import Runtime, create_runtime, gather_shares
from viff.comparison import Toft07Runtime
from viff.config import load_config
from viff.util import rand, find_prime

import pygtk
pygtk.require('2.0')
import gtk

import re
from math import log, floor
import time

class Protocol:
	"""Defining the protocol, which will be started at the
	bottom of the file."""

	def __init__(self, runtime):
		"""Imports the names of the participants and initiates
		the creation of the GUI."""

		# Save the Runtime for later use
		self.runtime = runtime

		# Read player names from config file
		file = open("author_config.txt", "r")
		lines = file.readlines()
		file.close()
		self.players = []
		for i in lines:
			# Transform unicode strings to UTF-8. Norwegian
			# letters would otherwise have caused warnings.
			self.players.append(re.sub("\n", "", str(i).decode("iso8859-15")))
		self.number_of_players = len(self.players)
		self.number_of_choices = len(self.players) - 1
		self.create_gui()
		
	def create_gui(self):
		"""The graphical interface is built using widgets from
		the PyGTK toolkit."""
		
		# Create main window		
		self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
		self.window.set_title("Rank the Authors")
		self.window.set_position(gtk.WIN_POS_CENTER)
		self.window.set_default_size(220, 100)
		self.window.set_border_width(10)

		# Create the outher box (widget container)
		vbox = gtk.VBox(False, 0)
		hbox = gtk.HBox(True, 0)
		self.window.add(vbox)
		vbox.pack_start(hbox, False, False, 0)

		# Table for arranging labels and radio buttons
		table = gtk.Table(self.number_of_players, self.number_of_players, False)
		table.set_row_spacings(2)
		table.set_col_spacings(10)
		j = 0
		for i in range(0, self.number_of_players):
			if not self.runtime.id == (i+1):
				# Name of players
				label = gtk.Label(self.players[i])
				table.attach(label, j+1, j+2, 0, 1)
				# 1th, 2th, 3th, etc. author
				label = gtk.Label(str(j+1) + "th")
				table.attach(label, 0, 1, j+1, j+2)
				j += 1

		# Set up grouped radio buttons. The exec statement is
		# used to create dynamic variable names.
		var_num = 0
		active = 1
		for i in range(0, self.number_of_choices):
			k = 0
			for j in range(0, self.number_of_players):
				name = str(j+1)
				if not self.runtime.id == (j+1):
					var_num += 1
					if (var_num % self.number_of_choices == 1):
						exec 'self.radio%s = self.setup_radio(None, "", name)' % var_num
					else:
						exec 'self.radio%s = self.setup_radio(self.radio%s, "", name)' % (var_num, var_num-1)
					exec 'table.attach(self.radio%s, k+1, k+2, i+1, i+2, 0)' % var_num
					k += 1
			# Setting some radio buttons as checked by default
			exec 'self.radio%s.set_active(True)' % active
			active += self.number_of_choices + 1
			
		vbox.pack_start(table, False, False, 0)
		
		# Separator
		separator = gtk.HSeparator()
		vbox.pack_start(separator, False, True, 10)

		# Submit button
		submitbox = gtk.HBox(False, 0)
		self.button = gtk.Button("Submit")
		self.button.connect("clicked", self.check_votes, None)
		submitbox.pack_start(self.button, True, False, 0)
		vbox.pack_start(submitbox, False, False, 0)

		# Separator	2	
		separator = gtk.HSeparator()
		vbox.pack_start(separator, False, True, 10)

		# Progress bar
		self.pbar = gtk.ProgressBar()
		vbox.pack_start(self.pbar, False, False, 0)

		# Separator	3	
		separator = gtk.HSeparator()
		vbox.pack_start(separator, False, True, 10)		

		# Output
		outputbox = gtk.HBox(False, 0)
		self.label = gtk.Label("Make your choice, " + str(self.players[self.runtime.id-1]) + "!")		
		outputbox.pack_start(self.label, True, False, 0)
		vbox.pack_start(outputbox, False, False, 0)

		self.window.show_all()

	def setup_radio(self, group, label, value):
		"""Function to simplify the display of grouped
		radio buttons."""
		
		radio = gtk.RadioButton(group, label)
		radio.set_name(value)
		return radio

	def on_error(self, widget):
		"""Function that displays an error message if
		check_votes	reports so."""
		
		md = gtk.MessageDialog(self.window, 
			gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, "Error:")
		md.format_secondary_text("You can only give one person one vote. Try again!")
		md.run()
		md.destroy()

	def check_votes(self, widget, data=None):
		"""Function for checking that the player have not given
		more than one vote for each person."""

		check = 1
		self.list = []
		buttion_number = 1
		for i in range(0, self.number_of_choices):
			exec 'group = self.radio%s.get_group()' % buttion_number in globals(), locals()
			for j in group:
				if j.get_active():
					for k in range(0, len(self.list)):
						if int(j.get_name()) == self.list[k]:
							check = 0
							self.on_error(None)
							break
					self.list.append(int(j.get_name()))
			buttion_number += self.number_of_choices
		if check == 1:
			self.submit(self, None)

	def submit(self, widget, data=None):
		"""Function for actions when pressing the
		submit button."""

		# Create a list of the points given. Player1's received
		# points is placed at points[0], player2's received
		# points at points[1], etc. First author gives n-1
		# points, second author gives n-2 points, etc.
		self.points = range(self.number_of_players)
		pts = self.number_of_choices
		for i in range(0, self.number_of_choices):
			x = self.list[i]
			self.points[x-1] = pts
			pts -= 1
		# Give yourself 0 points
		self.points[self.runtime.id-1] = 0

		# Create a temporary output text
		output = "For 1th author you selected: " + str(self.players[self.list[0]-1])
		self.label.set_text(output)
		self.button.set_sensitive(False)

		# Setup for the comparison protocol to work		
		l = self.runtime.options.bit_length
		k = self.runtime.options.security_parameter
		modulus = 2**65
		Zp = GF(find_prime(modulus, blum=True))

		# The input is secret shared. We end up with
		# n*n variables, where n = number of players
		self.time1 = time.clock()
		var_num = 0
		for i in range(0, self.number_of_players):
			player = 1
			for j in range(0, self.number_of_players):
				if self.runtime.id == player:
					exec 'a%s = self.runtime.shamir_share([player], Zp, self.points[i])' % var_num in globals(), locals()
				else:
					exec 'a%s = self.runtime.shamir_share([player], Zp)' % var_num in globals(), locals()
				var_num += 1
				player += 1

		self.shares = 0
		def step(value):
			"""Function for updating the progress bar."""
			
			self.shares += 1
			percent = round((self.shares*100.0) / float(self.number_of_players), 2)
			text = str(percent) + "%"
			self.pbar.set_fraction(self.shares / float(self.number_of_players))
			self.pbar.set_text(text)
			return value

		# Add callback for one variable per player. This way we
		# know when a user hits the submit button and we can
		# update the progress bar.
		for i in range(0, self.number_of_players):
			exec 'a%s.addCallback(step)' % i in globals(), locals()

		# Calculate the sum of points for each player
		array = []
		var_num = 0
		for i in range(0, self.number_of_players):
			exec 'sum%s = 0' % i in globals(), locals()
			for j in range(0, self.number_of_players):
				exec 'sum%s += a%s' % (i, var_num) in globals(), locals()
				var_num += 1
			exec 'array.append(sum%s)' % i in globals(), locals()

		# Initiate the sorting algorithm of the sums array
		sorted = self.sort(array)

		# Gather the results and call the self.results_ready
		# method when they have all been received
		sorted = gather_shares(map(self.runtime.open, sorted))
		sorted.addCallback(self.results_ready)
		
		# Open the sums (shares) and put them in a list
		list_of_sums = []
		for i in range(0, self.number_of_players):
			exec 'open%s = self.runtime.open(sum%s)' % (i, i) in globals(), locals()
			exec 'list_of_sums.append(open%s)' % i in globals(), locals()

		#sorted.addCallback(lambda _: self.runtime.shutdown())

	def sort(self, array):
		"""This sort function is provided as a part off the
		VIFF framework. The algorithm used is bitonic sort which
		was chosen in order to maximize the amount of work
		(comparisons) done in parallel."""

		# Make a shallow copy - the algorithm wont be
		# in-place anyway since we create lots of new
		# Shares as we go along.
		array = array[:]
		
		# Corresponding array for the order of authors according
		# to their points
		order = []
		for i in range(0, self.number_of_players):
			order.append(i+1)

		def bitonic_sort(low, n, ascending):
			if n > 1:
				m = n // 2
				bitonic_sort(low, m, ascending=not ascending)
				bitonic_sort(low + m, n - m, ascending)
				bitonic_merge(low, n, ascending)

		def bitonic_merge(low, n, ascending):
			if n > 1:
				# Choose m as the greatest power of 2
				# less than n.
				m = 2**int(floor(log(n-1, 2)))
				for i in range(low, low + n - m):
					compare(i, i+m, ascending)
				bitonic_merge(low, m, ascending)
				bitonic_merge(low + m, n - m, ascending)

		def compare(i, j, ascending):

			def xor(a, b):
				# TODO: We use this simple xor until
				# http://tracker.viff.dk/issue60 is fixed.
				return a + b - 2*a*b

			le = array[i] <= array[j] # a deferred

			# We must swap array[i] and array[j] when they
			# sort in the wrong direction, that is, when
			# ascending is True and array[i] > array[j], or
			# when ascending is False (meaning descending)
			# and array[i] <= array[j].
			#
			# Using array[i] <= array[j] in both cases we see
			# that this is the exclusive-or:
			b = xor(ascending, le)

			# We now wish to calculate
			#
			#   ai = b * array[j] + (1-b) * array[i]
			#   aj = b * array[i] + (1-b) * array[j]
			#
			# which uses four secure multiplications. We can
			# rewrite this to use only one secure multiplication
			ai, aj = array[i], array[j]
			b_ai_aj = b * (ai - aj)

			array[i] = ai - b_ai_aj
			array[j] = aj + b_ai_aj

			# Switch the order array correspondingly
			oi, oj = order[i], order[j]
			b_oi_oj = b * (oi - oj)				

			order[i] = oi - b_oi_oj
			order[j] = oj + b_oi_oj			

		bitonic_sort(0, len(array), ascending=False)
		return order

	def results_ready(self, results):
		"""This function is called as a callback above and the
		results variable will contain actual field elements, not
		just shares."""
		
		self.time2 = time.clock()
		self.time = self.time2 - self.time1
		print "time:", self.time
		
		# Unpack the list
		for i in range(0, self.number_of_players):
			exec 'r%s = int(results[i].value)' % i

		# Create the final output message
		output = ""
		for i in range(0, self.number_of_players):
			exec 'output += "%sth: " + self.players[r%s-1] + " "' % (i+1, i)

		self.label.set_text(output)
		print output
	
# Parse command line arguments.
parser = OptionParser()
Runtime.add_options(parser)
options, args = parser.parse_args()

if len(args) == 0:
    parser.error("You must specify a config file")
else:
    id, players = load_config(args[0])

# Create a deferred Runtime and ask it to run our protocol
# when ready
pre_runtime = create_runtime(id, players, 1, options, Toft07Runtime)
pre_runtime.addCallback(Protocol)

reactor.run()